import os
import sys
import time
import random
import numpy as np

from ase.io import read

from flare.atoms import FLARE_Atoms
from flare.bffs.sgp._C_flare import B1, B2, B3, NormalizedDotProduct
from flare.bffs.sgp.sparse_gp import SGP_Wrapper
from flare.bffs.sgp.calculator import SGP_Calculator
from flare.bffs.sgp.parallel_sgp import ParSGP_Wrapper
from copy import deepcopy

from mpi4py import MPI

np.random.seed(0)

def compute_mae(train_struc, dft_e, dft_f, dft_s, sgp_model):
    #train_struc.calc = SGP_Calculator(sgp_model)

    sgp_e = train_struc.get_potential_energy()
    sgp_f = train_struc.get_forces()
    sgp_s = train_struc.get_stress()

    mae_e = np.mean(np.abs(dft_e - sgp_e))
    mae_f = np.mean(np.abs(dft_f - sgp_f))
    mae_s = np.mean(np.abs(dft_s - sgp_s))

    mav_e = np.mean(np.abs(dft_e))
    mav_f = np.mean(np.abs(dft_f))
    mav_s = np.mean(np.abs(dft_s))

    mav_e_sgp = np.mean(np.abs(sgp_e))
    mav_f_sgp = np.mean(np.abs(sgp_f))
    mav_s_sgp = np.mean(np.abs(sgp_s))

    rel_e = 100 * mae_e / mav_e
    rel_f = 100 * mae_f / mav_f
    rel_s = 100 * mae_s / mav_s

    return mae_e, mav_e, mav_e_sgp, rel_e, mae_f, mav_f, mav_f_sgp, rel_f, mae_s, mav_s, mav_s_sgp, rel_s

def test(sgp_model, test_structures, g):
    pre_test = time.time()

    force_errors = []
    energy_errors = []
    stress_errors = []

    force_percents = []
    energy_percents = []
    stress_percents = []

    force_mags = []
    energy_mags = []
    stress_mags = []

    force_mags_sgp = []
    energy_mags_sgp = []
    stress_mags_sgp = []

    likelihood = sgp_model.likelihood

    # collect all dft results
    dft_energy = []
    dft_forces = []
    dft_stress = []
    for j, test_struc in enumerate(test_structures):
        dft_energy.append(test_struc.get_potential_energy())
        dft_forces.append(test_struc.get_forces())
        dft_stress.append(test_struc.get_stress())

    sgp_model.predict_on_structures(test_structures)
    for j, test_struc in enumerate(test_structures):
        dft_e = dft_energy[j]
        dft_f = dft_forces[j]
        dft_s = dft_stress[j]
        mae_e, mav_e, mav_e_sgp, rel_e, mae_f, mav_f, mav_f_sgp, rel_f, mae_s, mav_s, mav_s_sgp, rel_s = compute_mae(test_struc, dft_e, dft_f, dft_s, sgp_model)
        force_errors.append(mae_f)
        force_mags.append(mav_f)
        force_mags_sgp.append(mav_f_sgp)
        force_percents.append(rel_f)

        energy_errors.append(mae_e)
        energy_mags.append(mav_e)
        energy_mags_sgp.append(mav_e_sgp)
        energy_percents.append(rel_e)

        stress_errors.append(mae_s)
        stress_mags.append(mav_s)
        stress_mags_sgp.append(mav_s_sgp)
        stress_percents.append(rel_s)

    post_test = time.time() - pre_test
    if rank == 0:
        g.write(f"{len(sgp_model)}, {likelihood},"
            f" {1000*np.mean(force_errors)}, {1000*np.mean(force_mags)}, {100*np.mean(force_errors)/np.mean(force_mags)},"
            f" {1000*np.mean(energy_errors)}, {1000*np.mean(energy_mags)}, {100*np.mean(energy_errors)/np.mean(energy_mags)},"
            f" {1000*np.mean(stress_errors)}, {1000*np.mean(stress_mags)}, {100*np.mean(stress_errors)/np.mean(stress_mags)},"
            f" {post_test}, {total_opt_time}, {total_build_time}\n")
        g.flush()

N_max = int(sys.argv[1])
L_max = int(sys.argv[2])
cutoff = float(sys.argv[3])
power = int(sys.argv[4])

# -----------------------------------------------------------------------------
#                            Load training data.
# -----------------------------------------------------------------------------

train_structures  = read("element_cold_nequip_train.xyz", index=":")
train_structures += read("element_warm_nequip_train.xyz", index=":")
train_structures += read("element_melt_nequip_train.xyz", index=":")

train_structures = [FLARE_Atoms.from_ase_atoms(frame) for frame in train_structures]

test_structures   = read("element_cold_nequip_test.xyz", index=":")
test_structures  += read("element_warm_nequip_test.xyz", index=":")
test_structures  += read("element_melt_nequip_test.xyz", index=":")

# modify energies of all structures by subtracting mean energy
mean_energy = np.mean([frame.get_potential_energy() for frame in train_structures])
for frame in train_structures:
    frame.calc.results["energy"] -= mean_energy 

#test_structures = read("trj_test.xyz", index=":")
for frame in test_structures:
    frame.calc.results["energy"] -= mean_energy 

random.seed(42)
random.shuffle(train_structures)

# ------------------------------------------------------------------------------
#                         Prepare sparse GP model.
# -----------------------------------------------------------------------------

# Set up 2- and 3-body descriptors.
cutoff_name = "quadratic"
cutoff_hyps = []
# Set up many-body descriptors.
radial_basis = "chebyshev"

# -----------------------------------------------------------------------------
#                             Set up and Train Loop.
# -----------------------------------------------------------------------------

htms=[10,20,40,60,80,100,1000,2200,2700]

# MPI for writing to file
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
total_rank_size = comm.Get_size()
if rank == 0:
    g = open(f'learning_curve_K3_N{N_max}_L{L_max}_R{cutoff}_P{power}.txt','w')
else:
    g = None


#radial_hyps = [0, 9.0]
#settings = [len(species_map), 19, 0] 
#calc1 = B1(radial_basis, cutoff_name, radial_hyps, cutoff_hyps, settings)
#
#radial_hyps = [0, 5.0]
#settings = [len(species_map), 9, 4] 
#calc2 = B2(radial_basis, cutoff_name, radial_hyps, cutoff_hyps, settings)

radial_hyps = [0, cutoff]
species_map = {47:0}
settings = [len(species_map), N_max, L_max] 
calc3 = B2(radial_basis, cutoff_name, radial_hyps, cutoff_hyps, settings)

calc_list = [calc3] #[calc1, calc2, calc3]

# Define kernel for each descriptor
#sigma1 = 10
#kernel1 = NormalizedDotProduct(sigma1, power)

sigma2 = 3
kernel2 = NormalizedDotProduct(sigma2, power)

#sigma3 = 0.5
#kernel3 = NormalizedDotProduct(sigma3, power)

kernel_list = [kernel2] #[kernel1, kernel2, kernel3] 

sigma_e = 0.1
sigma_f = 0.1
sigma_s = 0.01
max_iterations = 100

#sparse_gp = SGP_Wrapper(
sparse_gp = ParSGP_Wrapper(
    kernel_list,
    calc_list,
    cutoff,
    sigma_e,
    sigma_f,
    sigma_s,
    species_map,
    variance_type="local",
    energy_training=True,
    force_training=True,
    stress_training=True,
    max_iterations=max_iterations,
)
  
learning_curve = []
hyp_opt_moments = []
hyp_gradient = []
hyps = []

total_opt_time = 0
if rank == 0:
    g.write("frame, likelihood, force_mae (meV/A), force_mav_dft (meV/A), force_percent, energy_mae (meV), energy_mav_dft (meV), energy_percent, stress_mae (meV/A^3), stress_mav_dft (meV/A^3), stress_percent, time_test, cum_hyp_opt, time_build\n")

# training and testing
sparse_size = 1
total_build_time = 0
for i, train_struc in enumerate(train_structures):
    pre_build = time.time()
    if len(sparse_gp) == 0:
        if sparse_size >= total_rank_size:
            init_n_envs = sparse_size
        else:
            init_n_envs = total_rank_size
        random_indices = np.random.choice(len(train_struc), min(len(train_struc), init_n_envs), replace=False).tolist()

        sparse_gp.update_db(
    	   train_struc,
    	   custom_range=[random_indices for k in range(len(kernel_list))],
    	   mode="specific",
        )

    else:
        custom_range = [np.random.choice(len(train_struc), replace=False, size=min(sparse_size, len(train_struc))).tolist()]
        sparse_gp.update_db(
    	   train_struc,
    	   custom_range=[min(sparse_size, len(train_struc))] * len(kernel_list),
           #custom_range=custom_range,
    	   mode="uncertain",
           #mode="specific",
        )

    total_build_time += time.time() - pre_build

    pre_opt = time.time()
    if len(sparse_gp) in htms:
        sparse_gp.train()
        print("SGP size", len(sparse_gp))
        print("Hyperparameters", sparse_gp.hyps)
        print("Likelihood Grad", sparse_gp.likelihood_gradient)
        total_opt_time += time.time() - pre_opt
        # run testing
        test_structures_new = [deepcopy(struc) for struc in test_structures]
        test(sparse_gp, test_structures_new, g)
    
# run testing
test_structures_new = [deepcopy(struc) for struc in test_structures]
test(sparse_gp, test_structures_new, g)

sparse_gp.write_model(name="Ag_2700cwm")

if rank == 0:
    g.close()

